@@ -1,4 +1,4 @@
-# flake8: NOQA W605
+# flake8: NOQA: W605
 """Pythonization of the :ref:`tmux(1)` pane.
 
 libtmux.pane
@@ -14,8 +14,7 @@ from typing import overload
 from libtmux.common import tmux_cmd
 from libtmux.neo import Obj, fetch_obj
 
-from . import exc, formats
-from .common import PaneDict
+from . import exc
 
 if t.TYPE_CHECKING:
     from typing_extensions import Literal
@@ -103,9 +102,10 @@ class Pane(Obj):
     def session(self) -> "Session":
         return self.window.session
 
-    #
-    # Command (pane-scoped)
-    #
+    """
+    Commands (pane-scoped)
+    """
+
     def cmd(self, cmd: str, *args: t.Any, **kwargs: t.Any) -> tmux_cmd:
         """Return :meth:`Server.cmd` defaulting to ``target_pane`` as target.
 
@@ -115,13 +115,14 @@ class Pane(Obj):
         ``args`` will override using the object's ``pane_id`` as target.
         """
         if not any(arg.startswith("-t") for arg in args):
-            args = ("-t", self.pane_id) + args
+            args = ("-t", self.pane_id, *args)
 
         return self.server.cmd(cmd, *args, **kwargs)
 
-    #
-    # Commands (tmux-like)
-    #
+    """
+    Commands (tmux-like)
+    """
+
     def resize_pane(self, *args: t.Any, **kwargs: t.Any) -> "Pane":
         """
         ``$ tmux resize-pane`` of pane and return ``self``.
@@ -155,13 +156,42 @@ class Pane(Obj):
         self.refresh()
         return self
 
-    def capture_pane(self) -> t.Union[str, t.List[str]]:
+    def capture_pane(
+        self,
+        start: t.Union["Literal['-']", t.Optional[int]] = None,
+        end: t.Union["Literal['-']", t.Optional[int]] = None,
+    ) -> t.Union[str, t.List[str]]:
         """
         Capture text from pane.
 
         ``$ tmux capture-pane`` to pane.
+        ``$ tmux capture-pane -S -10`` to pane.
+        ``$ tmux capture-pane`-E 3` to pane.
+        ``$ tmux capture-pane`-S - -E -` to pane.
+
+        Parameters
+        ----------
+        start: [str,int]
+            Specify the starting line number.
+            Zero is the first line of the visible pane.
+            Positive numbers are lines in the visible pane.
+            Negative numbers are lines in the history.
+            `-` is the start of the history.
+            Default: None
+        end: [str,int]
+            Specify the ending line number.
+            Zero is the first line of the visible pane.
+            Positive numbers are lines in the visible pane.
+            Negative numbers are lines in the history.
+            `-` is the end of the visible pane
+            Default: None
         """
-        return self.cmd("capture-pane", "-p").stdout
+        cmd = ["capture-pane", "-p"]
+        if start is not None:
+            cmd.extend(["-S", str(start)])
+        if end is not None:
+            cmd.extend(["-E", str(end)])
+        return self.cmd(*cmd).stdout
 
     def send_keys(
         self,
@@ -248,12 +278,13 @@ class Pane(Obj):
         self.cmd("display-message", cmd)
         return None
 
-    #
-    # Commands ("climber"-helpers)
-    #
-    # These are commands that climb to the parent scope's methods with
-    # additional scoped window info.
-    #
+    """
+    Commands ("climber"-helpers)
+
+    These are commands that climb to the parent scope's methods with
+    additional scoped window info.
+    """
+
     def select_pane(self) -> "Pane":
         """
         Select pane. Return ``self``.
@@ -265,7 +296,7 @@ class Pane(Obj):
         assert isinstance(self.pane_id, str)
         pane = self.window.select_pane(self.pane_id)
         if pane is None:
-            raise exc.LibTmuxException(f"Pane not found: {self}")
+            raise exc.PaneNotFound(pane_id=self.pane_id)
         return pane
 
     def split_window(
@@ -297,9 +328,10 @@ class Pane(Obj):
             percent=percent,
         )
 
-    #
-    # Commands (helpers)
-    #
+    """
+    Commands (helpers)
+    """
+
     def set_width(self, width: int) -> "Pane":
         """
         Set width of pane.
@@ -352,7 +384,7 @@ class Pane(Obj):
         return self.pane_id == other.pane_id
 
     def __repr__(self) -> str:
-        return "{}({} {})".format(self.__class__.__name__, self.pane_id, self.window)
+        return f"{self.__class__.__name__}({self.pane_id} {self.window})"
 
     #
     # Aliases
@@ -412,12 +444,12 @@ class Pane(Obj):
         """
         .. deprecated:: 0.16
         """
-        warnings.warn("Pane.get() is deprecated")
+        warnings.warn("Pane.get() is deprecated", stacklevel=2)
         return getattr(self, key, default)
 
     def __getitem__(self, key: str) -> t.Any:
         """
         .. deprecated:: 0.16
         """
-        warnings.warn(f"Item lookups, e.g. pane['{key}'] is deprecated")
+        warnings.warn(f"Item lookups, e.g. pane['{key}'] is deprecated", stacklevel=2)
         return getattr(self, key)
